// ///////////////////////////////////////////////////////////////////
// ------------------------------------------------------------------
// COPYRIGHT (c) 2004-2016 <VirtualSim>
// <legal statement>
// ------------------------------------------------------------------
//
// Software Unit Name: //&&
//
// ------------------------------------------------------------------
// All rights reserved.
// http://www.virtualsim.com
// ///////////////////////////////////////////////////////////////////

#ifndef VT_BASIC_DYN_H
#define VT_BASIC_DYN_H

// [!VT]  DEFINITION
// ------------------
// Include directives
// ------------------

#include <vt_components.h>
#include "coords.h"

// --------------------
// Forward declarations
// --------------------

// Local includes

// -----------------
// Class declaration
// -----------------
// [!VT]  END

// [!VT] HELP
/*! \brief BasicDyn component class

See the documentation (click the button or refer to vsTaskerDevGuide)
*/
// [!VT]  END

// Useful defines

#define KMH2MS(s) (s/3.6)
#define MS2KMH(s) (s*3.6)

//! CrashData used to store specific data at crash time
//! This structure is normally private.
typedef struct {
   double  rot_v; 
   double  rot_i; 
   double  spd_v; 
   double  hdg_i;
   int     mode;
} CrashData;

//! AltMode specifies the way the model shall reach the target altitude.
typedef enum {
   AP_Altitude,   //!< target altitude is reached and maintained
   AP_Slope,      //!< Slope to be maintained disregarding target altitude
   AP_Alt_Slope   //!< Slope to be maintained until target altitude reached, AP_Altitude then
} AltMode;

//! CtrlMode specifies the how attitude values must be changed by the dynamic.
typedef enum {
   AP_Auto,       //!< Auto-pilot, pitch and roll change according to alt and yaw
   AP_Joystick    //!< Joystick control, pitch and roll change according to joystick data
} CtrlMode;

//! DynState specifies the current state of the dynamic (model)
typedef enum {
   ST_Normal     = 0, //!< default mode
   ST_Airborne   = 0, //!< alias of normal
   ST_Grounded   = 1, //!< mostly used by air entities
   ST_Landed     = 1, //!< alias of grounded
   ST_Crashed    = 2, //!< entity is crashed or wrecked
   ST_Wreck      = 2, //!< alias of crashed
} DynState;

// [!VL]  MODEL_ONLY    true		
// [!VL]  INSTANCIATE   false


//====================================================
class BasicDyn : public Vt_Component
{
// [!VT]  CLASS

// Methods section:
public:
  
  //! must be called before used by the Entity holding the component.
  //! normal use: set(entity->getPos());
  //! \param pointer to the Entity coordinate object
  void setPos(WCoord* pos);  
  void set(WCoord* pos) { setPos(pos); }  //!< alias of setPos(), deprecated

  //! use this function to offset the current position (like for wind)
  //! this function will keep the velocity vector of high fidelity cinematic components
  virtual void shiftPos(float x, float y, float z);

  //! return position set above (normally, entity position)
  WCoord* getPos();
  			 
  //! force altitude down to the ground
  virtual void clampZ()=0;     

  //! returns the time in seconds to reach the given position and the current speed
  //! This does not take into account the heading time change. 
  //! \param coordinates in XYZ type
  virtual double timeTo(WCoord&)=0;
  virtual double timeTo2D(WCoord&)=0;   //!< same as timeTo() but discard the elevation. Only the ground projected position matters.

  //! \return the position expected after the given number of second
  //! \param time value in seconds
  virtual WCoord& getPosIn(float time);

  //! Initiate Crash behavior. Depending of the Entity type, behavior will vary.
  virtual void startCrash();
  
  //! Returns the current Crash status
  //! \return 1 if crashed and 0 otherwise
  virtual int  isCrashed();
  
  //! Returns the current Crash status
  //! \return 1 if not crashed, 0 otherwise
  virtual int  isNotCrashed();

  //! Return if great-circle must be used for move() function
  //! \return true if great-circle to be used, false otherwise (default)
  int useGC();

  //! Force the great-circle mode ON or OFF
  void useGC(int mode);

  virtual void stabilize();				  //!< make the entity steady, flat, no acceleration nor rotation
  virtual void freeze();				  //!< freeze the motion
  virtual void release();				  //!< release the motion

  void  controlMode(CtrlMode);			  //!< See enum CtrlMode
  virtual void altitudeMode(AltMode, float slope =0);	//!< See enum AltMode

  //! Switch ON or OFF the auto-pilot mode (if supported by the model).
  //! Call controlMode with the appropriate enum
  //! \param ON or OFF
  void switchAutoPilot(int);

  // Basic API
  virtual void setSpeed(double)=0;        //!< set the target speed in meters per seconds
  virtual void setHeading(double)=0;      //!< set the target heading, in radians
  virtual void setHeading(WCoord&)=0;     //!< set the heading so that to face the given position x,y
  virtual void setAltitude(double)=0;     //!< set the target altitude in meters
  virtual void setAltitude(WCoord&)=0;    //!< set the target altitude as the z value of the parameter
  virtual void setElevation(double)=0;    //!< set the target elevation (slope) in radians
  virtual void setElevation(WCoord&)=0;   //!< set the target elevation in order to reach the given z value of the parameter
  virtual void setRotation(double)=0;     //!< set the target rotation in radians (0 = flat, 180 = upside down)

  virtual void forceAltitude(double)=0;   //!< force the current altitude in meters
  virtual void forceHeading(double)=0;    //!< force the current azimuth in radians
  virtual void forceHeading(WCoord&)=0;   //!< force the current azimuth so that to face the given position x,y
  virtual void forceSpeed(double)=0;      //!< force the current speed in meters per seconds

  virtual void forceElevation(double) {}  //!< force the current azimuth in radians
  virtual void forceElevation(WCoord&) {} //!< force the current azimuth so that to face the given position z
  
  virtual void incSpeed(double)=0;        //!< increment the target speed with the parameter value in m/s
  virtual void incHeading(double)=0;      //!< increment the target heading with the parameter value in radians

  virtual double getBearingTo(WCoord&)=0; //!< returns the relative bearing in radians to get there 
  virtual double getSiteTo(WCoord&)=0;    //!< returns the relative elevation in radians to get there 

  virtual void incAltitude(double)=0;     //!< increment the target altitude. \param increment value in meters
  virtual void incElevation(double)=0;    //!< increment the target elevation (slope). \param increment value in radian
  virtual void incRotation(double)=0;     //!< increment the target rotation. \param increment value in radian

  virtual int speedOk(float =2);		  //!< Returns true if |current speed - requested speed| < param (m/s)
  virtual int headingOk(float =3);		  //!< Returns true if |current heading - requested heading| < param (deg)
  virtual int altitudeOk(float =5);		  //!< Returns true if |current altitude - requested altitude| < param (m)

  virtual void   setDefSpeed(double);     //!< \param speed in meters per seconds
  virtual double getDefSpeed();           //!< \return default speed in meters per seconds

  void setAlt(double);					  //!< alias of setAltitude
  void setAlt(WCoord&);				      //!< alias of setAltitude
  void setSpd(double);					  //!< alias of setSpeed
  void setHdg(double);					  //!< alias of setHeading
  void setHdg(WCoord&);					  //!< alias of setHeading
  void setYaw(double);					  //!< alias of setHeading
  void setYaw(WCoord&);					  //!< alias of setHeading
  void setAzim(double);					  //!< alias of setHeading
  void setAzim(WCoord&);				  //!< alias of setHeading
  void setElev(double);					  //!< alias of setElevation
  void setElev(WCoord&);				  //!< alias of setElevation
  void setPitch(double);				  //!< alias of setElevation
  void setPitch(WCoord&);				  //!< alias of setElevation
  void setRoll(double);					  //!< alias of setRotation

  void incSpd(double);					  //!< alias of incSpeed
  void incAlt(double);					  //!< alias of incAltitude
  void incHdg(double);					  //!< alias of incHeading
  void incYaw(double);					  //!< alias of incHeading
  void incAzim(double);					  //!< alias of incHeading
  void incElev(double);					  //!< alias of incElevation
  void incPitch(double);				  //!< alias of incElevation
  void incRoll(double);					  //!< alias of incRotation

  virtual void setBrake(float);			  //!< activate the brake to reduce the speed. \param 0..1 0=None, 1=full
  virtual float getBrake();				  //!< \return current brake factor [0..1]

  virtual double getSpeed()=0;			  //!< \return current speed (in m/s)
  virtual double getHeading()=0;		     //!< \return current heading/azimuth (in radian)
  virtual double getAltitude()=0;		  //!< \return current altitude (in meters)
  virtual double getElevation()=0;		  //!< \return current elevation (in radian)
  virtual double getRotation()=0;		  //!< \return current rotation (in radian)

  virtual double getTargetSpeed()=0;      //!< \return target speed (in m/s)
  virtual double getTargetHeading()=0;    //!< \return target heading/azimuth (in radian)
  virtual double getTargetAltitude()=0;   //!< \return target altitude (in meters)
  virtual double getTargetElevation()=0;  //!< \return target elevation (in radian)
  virtual double getTargetRotation()=0;   //!< \return target rotation (in radian)

  double getSpd();						  //!< alias of getSpeed
  double getAlt();						  //!< alias of getAltitude
  double getYaw();						  //!< alias of getHeading
  double getHdg();						  //!< alias of getHeading
  double getAzim();						  //!< alias of getHeading
  double getElev();						  //!< alias of getElevation
  double getPitch();					     //!< alias of getElevation
  double getRoll();						  //!< alias of getRotation

  double getTargetSpd();				  //!< alias of getTargetSpeed
  double getTargetAlt();				  //!< alias of getTargetAltitude
  double getTargetHdg();				  //!< alias of getTargetHeading
  double getTargetYaw();				  //!< alias of getTargetHeading
  double getTargetAzim();				  //!< alias of getTargetHeading
  double getTargetElev();				  //!< alias of getTargetElevation
  double getTargetPitch();				  //!< alias of getTargetElevation
  double getTargetRoll();				  //!< alias of getTargetRotation

  //void   setHAT(double h);			  //!< Set Height Above Terrain. Default = 0
  //double getHAT();						  //!< Deprecated

  void setState(DynState);
  DynState getState();

  //! Set the velocity vector (for effectiveness use processVelocity below)
  //! \param vv is the velocity vector in m/s
  void setVelocity(Vec3D& vv);

  //! Set the velocity vector (for effectiveness use processVelocity below)
  //! \param 1 x axis speed value in m/s
  //! \param 2 y axis speed value in m/s
  //! \param 3 z axis speed value in m/s
  void setVelocity(float, float, float);

  //! Change the speed, heading and pitch (target) according to the velocity vector
  //! Use this function after setVelocity.
  void processVelocity();

  //! Force the speed, heading and pitch (values) according to the velocity vector
  //! Use this function after setVelocity.
  void forceVelocity();

  //! Return the current velocity vector. Use this function after computeVelocity() as the vector
  //! is not automatically updated (for performance reason).
  Vec3D& getVelocity(); 

  //! Compute the velocity vector based on the current speed, azimuth, heading and pitch
  void computeVelocity(float yaw, float pitch, float speed);
  virtual void computeVelocity();
 
//! \brief Variables
//! 

  // Data section:
public: 
   AltMode    altitude_mode;   //&&  LBL[Altitude Mode] DEF[AP_Altitude] HINT[Altitude Mode] LIST[AP_Altitude,AP_Slope,AP_Alt_Slope] SET[altitudeMode]
   CtrlMode   control_mode;    //&&  LBL[Control Mode]  DEF[AP_Auto]     HINT[Autopilot or Joystick control] LIST[AP_Auto,AP_Joystick] SET[controlMode]
   double     def_spd;         //&&  LBL[Default Speed]  DEF[10] HINT[Normal speed to go]  SET[setDefSpeed]  UNIT[m/s]
   float      brake;           //&&  LBL[Brake Value] DEF[0] HINT[Braking from 0 (no brake) to 1 (full brake)]

protected:
   DynState   state;           //&&  LBL[State] DEF[Normal] LIST[Normal=ST_Normal, Grounded=ST_Grounded, Crashed=ST_Crashed] HINT[Current State of the Dynamic Mode]
   int        use_gc;          //&&  LBL[Use Great-Circle] LIST[No=0,Yes=1] DEF[No] HINT[Use great-circle computation for move()]

   int        frozen;          // used only in move()
   Vec3D      velocity;        // velocity vector, set or computed
   double     hat_deprecated;  // use pft_body HAT instead, deprecated
   CrashData  crash;           // depending of the model, change the dynamic in crash mode
   WCoord*    pos;             // position of the dynamic model center

protected:
  //! set crash.mode to 2
  //! raise _Crashed event to entity (only once)
  //! set PtfStatus damage to _Wreck (if can do)
  virtual void setCrashed();


// [!VT]  END
public:
	BasicDyn(int id);

	virtual int  tic();
	virtual void setUserCode();
	virtual void setDefault();
	virtual int  react(int, MData*);
	virtual void setUserData(UsrInterface*);
	virtual void update(char* =0);

private:

};

inline void    BasicDyn::setPos(WCoord* p) { pos = p; }  
inline WCoord* BasicDyn::getPos() { return pos; }
inline void    BasicDyn::shiftPos(float x, float y, float z) { pos->x += x, pos->y +=y, pos->z += z; }

inline void BasicDyn::setSpd(double spd)    { setSpeed(spd); }
inline void BasicDyn::setAlt(double alt)    { setAltitude(alt); }
inline void BasicDyn::setYaw(double hdg)    { setHeading(hdg); }
inline void BasicDyn::setHdg(double hdg)    { setHeading(hdg); }
inline void BasicDyn::setAzim(double azm)   { setHeading(azm); }
inline void BasicDyn::setElev(double elv)   { setElevation(elv); }
inline void BasicDyn::setPitch(double elv)  { setElevation(elv); }
inline void BasicDyn::setRoll(double rol)   { setRotation(rol); }

inline void BasicDyn::setAlt(WCoord& pos)   { setAltitude(pos); }
inline void BasicDyn::setHdg(WCoord& pos)   { setHeading(pos); }
inline void BasicDyn::setYaw(WCoord& pos)   { setHeading(pos); }
inline void BasicDyn::setAzim(WCoord& pos)  { setHeading(pos); }
inline void BasicDyn::setElev(WCoord& pos)  { setElevation(pos); }
inline void BasicDyn::setPitch(WCoord& pos) { setElevation(pos); }

inline void BasicDyn::incSpd(double spd)    { incSpeed(spd); }
inline void BasicDyn::incAlt(double alt)    { incAltitude(alt); }
inline void BasicDyn::incHdg(double hdg)    { incHeading(hdg); }
inline void BasicDyn::incYaw(double hdg)    { incHeading(hdg); }
inline void BasicDyn::incAzim(double azm)   { incHeading(azm); }
inline void BasicDyn::incElev(double elv)   { incElevation(elv); }
inline void BasicDyn::incRoll(double rol)   { incRotation(rol); }

inline double BasicDyn::getSpd()   { return getSpeed(); }
inline double BasicDyn::getAlt()   { return getAltitude(); }
inline double BasicDyn::getHdg()   { return getHeading(); }
inline double BasicDyn::getYaw()   { return getHeading(); }
inline double BasicDyn::getAzim()  { return getHeading(); }
inline double BasicDyn::getElev()  { return getElevation(); }
inline double BasicDyn::getPitch() { return getElevation(); }
inline double BasicDyn::getRoll()  { return getRotation(); }

inline void   BasicDyn::setDefSpeed(double val) { def_spd = val; }
inline double BasicDyn::getDefSpeed() { return def_spd; }

inline void   BasicDyn::setBrake(float brk) { brake = brk; }
inline float  BasicDyn::getBrake()			{ return brake; }

inline double BasicDyn::getTargetSpd()   { return getTargetSpeed(); }
inline double BasicDyn::getTargetAlt()   { return getTargetAltitude(); }
inline double BasicDyn::getTargetHdg()   { return getTargetHeading(); }
inline double BasicDyn::getTargetYaw()   { return getTargetHeading(); }
inline double BasicDyn::getTargetAzim()  { return getTargetHeading(); }
inline double BasicDyn::getTargetElev()  { return getTargetElevation(); }
inline double BasicDyn::getTargetPitch() { return getTargetElevation(); }
inline double BasicDyn::getTargetRoll()  { return getTargetRotation(); }

inline int BasicDyn::speedOk(float)      { return 1; }
inline int BasicDyn::headingOk(float)    { return 1; }
inline int BasicDyn::altitudeOk(float)   { return 1; }

inline void BasicDyn::switchAutoPilot(int on) { if (on) controlMode(AP_Auto); else controlMode(AP_Joystick); }
inline void BasicDyn::controlMode(CtrlMode mode) { control_mode = mode; }
inline void BasicDyn::altitudeMode(AltMode mode, float slope) { altitude_mode = mode; }

inline int  BasicDyn::isCrashed()    { return (crash.mode == 2); }
inline int  BasicDyn::isNotCrashed() { return !isCrashed(); }

//inline void   BasicDyn::setHAT(double h) { hat = h; }
//inline double BasicDyn::getHAT()  { return hat; }

inline void BasicDyn::freeze()  { frozen = 1; }
inline void BasicDyn::release() { frozen = 0; }

inline DynState BasicDyn::getState() { return state; }
inline void BasicDyn::setState(DynState st) { state = st; }

inline void BasicDyn::setVelocity(Vec3D& vv) { velocity = vv; }
inline void BasicDyn::setVelocity(float vx, float vy, float vz) { velocity.set(vx, vy, vz); }

inline Vec3D& BasicDyn::getVelocity()   { return velocity; } 
inline void   BasicDyn::useGC(int mode) { use_gc = mode; }
inline int    BasicDyn::useGC()          { return use_gc; }

#endif
